LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_arith.ALL;
USE ieee.std_logic_unsigned.ALL;
USE work.controlFuncs.ALL;

ENTITY controller IS
  
  GENERIC (
    nPlayers : INTEGER := 3);          -- Number of Players

  PORT (
    -- Input
    clock       : IN  STD_LOGIC;
    enable      : IN  STD_LOGIC;
    reset       : IN  STD_LOGIC;
    bistEnable  : IN  STD_LOGIC;
    -- Input from Player
    pHit        : IN  STD_LOGIC_VECTOR(nPlayers-1 DOWNTO 0);
    pStand      : IN  STD_LOGIC_VECTOR(nPlayers-1 DOWNTO 0);
    -- Input from Card Generator
    issued      : IN  STD_LOGIC;
    -- Input from Bank
    gameOver    : IN  STD_LOGIC;
    balanceRdy  : IN  STD_LOGIC;
    -- Input from Score Keeper
    skBust      : IN  STD_LOGIC;
    -- Output
    playerID    : OUT STD_LOGIC_VECTOR(nPlayers-1 DOWNTO 0);
    globalReset : OUT STD_LOGIC;
    -- Output to Card Generator
    cgIssue     : OUT STD_LOGIC;
    cgNewHand   : OUT STD_LOGIC;
    cgBist      : OUT STD_LOGIC;
    -- Output to Score Keeper
    handOver    : OUT STD_LOGIC;
	-- Output to Card Generator
	shuffle		: OUT STD_LOGIC
	);


END controller;

ARCHITECTURE mixed OF controller IS

  SIGNAL state    : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0000";
  SIGNAL numCards : STD_LOGIC                    := '0';
  
  SIGNAL forced_reset: STD_LOGIC; -- This reset is used by the controller to forcibly stop the hand when a timeout is encountered.
 
  SIGNAL count_card: STD_LOGIC_VECTOR(7 downto 0);
BEGIN

  main : PROCESS (clock, reset, issued)
    VARIABLE initialPlayer : STD_LOGIC_VECTOR(nPlayers-1 DOWNTO 0) := (OTHERS => '0');
    VARIABLE controlID     : STD_LOGIC_VECTOR(nPlayers-1 DOWNTO 0) := (OTHERS => '0');
    VARIABLE timeout       : STD_LOGIC_VECTOR(8 DOWNTO 0); -- 256 clock ticks to think!
  BEGIN 
  -- process main
  -- Reset condition initial playerID to be set to the first player.
     IF reset = '1' or forced_reset='1' THEN  -- asynchronous reset (active low)
      initialPlayer := (OTHERS => '0');
      initialPlayer(initialPlayer'right) := '1';
      playerID <= controlID;
	  count_card <= (OTHERS => '0');
	  forced_reset<='0';
      globalReset <= '1';
      state <= "0000";
      cgBist <= '0';
      handOver <= '0';
	  shuffle <='1'; -- In case of total reset, shuffle the cards once
    ELSIF clock'event AND clock = '1' AND reset = '0' THEN  -- rising clock edge
      globalReset <= '0'; 
      cgBist <= '0';
	  if bistEnable = '1' then
	  cgBist<='1';
	  else 
	  cgBist<='0';
	  end if;

      CASE state IS
	  -- The state 0000 is the initial state of the game when a new hand is initiated. The cgNewHand signal is set to begin the game.
        WHEN "0000" =>                  -- New Hand Init
          cgNewHand <= '1';
		  shuffle <='0';
          playerID  <= initialPlayer; -- Assign control to the first player
          controlID := initialPlayer;
          state     <= "0001"; -- Move on with the game ...go to bet mode.
          timeout   := (OTHERS => '0');
-- This state is the second stage where the players place their respective bets.This is a precursor to the issual of cards to the players. if a player fails to place a bet on time, the hand is abandoned
        WHEN "0001" =>                  -- Bet mode
          cgNewHand <= '0';
          numCards  <= '0';

          -- Cycle Through Players
          -- If pStand or timeout ( once bet is made, the player sets a stand signal...so wait for stand or timeout whichever happens first )
          
          IF (pStand AND controlID) = controlID OR timeout(timeout'left) = '1' THEN
		 -- if bet was placed on time, move on to next player	
			if (pStand AND controlID) = controlID THEN
            	controlID := nextPlayerID(controlID);
            	playerID  <= controlID;
                --ASSERT false REPORT "PlayerID Set" SEVERITY note;            
            	timeout   := (OTHERS => '0');
            	IF controlID = initialPlayer THEN
              		state <= "0010";
            	END IF;
			else 
		  -- This is the timeout condition ==> stop the hand!
			  state<="0101"; -- set the state to game(of current hand)over state
			   ASSERT state="0101"
			   REPORT "Timed out on a player's bet....stopping game"
			   SEVERITY note;
			end if;
          ELSE
            --   if bet is yet to be placed, increment timeout;
            timeout := timeout + 1;
-- Are the next 2 lines needed?? 
--            ASSERT timeout(timeout'left) = '0'
--			 REPORT "Time Out!!! on player " & integer'image(conv_integer(controlID))  SEVERITY note;            
          END IF;
        WHEN "0010" =>                  -- Deal Mode
          --  Cycle Through Players
          cgIssue <= '1';
          IF issued = '1' THEN
            cgIssue   <= '0';
            controlID := nextPlayerID(controlID); -- cycle through players
            IF controlID(controlID'left) = '1' THEN
              controlID := nextPlayerID(controlID); -- if the new player generated is the dealer then skip him because his cards are not dealt in this mode.
            END IF;
            playerID <= controlID;
            IF controlID = initialPlayer THEN
              timeout := (OTHERS => '0');
              IF numCards = '1' THEN
                state   <= "0011";
				count_card<=count_card+((nPlayers-1)*2 + 1);
				-- Keeping track of cards dealt so that appropriate shuffling can be requested from the card generator
              ELSE
                numCards <= '1';
              END IF;

            END IF;
          END IF;
          
        WHEN "0011" =>                         -- Game Interaction
          -- For each player
          -- When player is enabled their hit/stand must be preset (i.e. 0's if
          -- they don't want to do anything yet). 
          IF skBust = '1' OR ((pStand AND controlID) = controlID) OR timeout(timeout'left) = '1' THEN
            timeout   := (OTHERS => '0');
            cgIssue   <= '0';
            controlID := nextPlayerID(controlID);
            IF controlID(controlID'left) = '1' THEN
              controlID := nextPlayerID(controlID);
            END IF;
            playerID <= controlID;
            IF controlID = initialPlayer THEN  -- service dealer
              controlID                 := (OTHERS => '0');
              controlID(controlID'left) := '1';
              playerID                  <= controlID;
              state                     <= "0110";
            END IF;
          ELSE
            timeout := timeout + 1;
            IF (pHit AND controlID) = controlID THEN
              cgIssue <= '1';
			  count_card<=count_card+1;
            ELSE
              cgIssue <= '0';
            END IF;

            IF issued = '1' THEN
              timeout := (OTHERS => '0');
            END IF;

          END IF;
          
          
        WHEN "0110" =>
          -- deal with dealer
          IF skBust = '1' OR ((pStand AND controlID) = controlID) OR timeout(timeout'left) = '1' THEN
            timeout   := (OTHERS => '0');
            cgIssue   <= '0';
            controlID := initialPlayer;
            playerID  <= controlID;
            handOver <= '1';
            state     <= "0100";

          ELSE
            timeout := timeout + 1;
            IF (pHit AND controlID) = controlID THEN
              cgIssue <= '1';
			  count_card<=count_card+1;
            ELSE
              cgIssue <= '0';
            END IF;

            IF issued = '1' THEN
              timeout := (OTHERS => '0');
            END IF;

          END IF;
          
        WHEN "0100" =>                  -- Results
          -- Cycle through each player
          IF balanceRdy = '1' THEN
            controlID := nextPlayerID(controlID);
            playerID  <= controlID;

            IF controlID = initialPlayer THEN
              handOver <= '0';
              playerID <= (OTHERS => '0');
              state <= "0101";
            END IF;
            
          END IF;

        WHEN "0101" =>                  -- Game Over

          IF gameOver = '1' THEN
            -- display results from banker
			forced_reset<='1';
            ELSE
		    if count_card >= "001101000" then  -- if cards dealt is 104 then ask for a shuffle
				shuffle<='1';
			  end if;
              state <= "0000";
          END IF;
        WHEN OTHERS => NULL;
      END CASE;
      
    ELSE
      globalReset <= '0'; 
    END IF;
  END PROCESS main;

END mixed;


